#include <iostream>
#include <cstdlib>
using std::cout;
using std::endl;

#include "game_jam_game.hpp"
#include "input.hpp"
#include "render.hpp"
#include "texture.hpp"

float const fHalfDungeonWidth = 10.f / 2.f;
float const fHalfDungeonHeight = 5.5f / 2.f;

GameJamGame::GameJamGame()
: state (TitleScreen)
{
    /**
     * Populate cache.
     */

    Reload();
}

StaticMesh GameJamGame::LoadStaticMesh(std::string const & sMeshName,
                                              std::string const & sMeshFile,
                                              std::string const & sTextureFile,
                                              uint32_t textureWidth,
                                              uint32_t textureHeight)
{
    cout << "Loading static mesh: " << sMeshName << " ( " << sMeshFile << ", " << sTextureFile << ")..." << endl;

    auto it = staticMeshes.find(sMeshName);

    if(it != staticMeshes.end())
    {
        cout << sMeshName << " was cached" << endl;
        return it->second;
    }

    cout << sMeshName << " was not cached - loading.";

    StaticMesh sm = StaticMesh::LoadFromFile(sMeshFile, sTextureFile, 512, 512);
    staticMeshes[sMeshName] = sm;
    return sm;
}

void * GameJamGame::LoadTexture(std::string const & sTextureFile,
                                uint32_t width,
                                uint32_t height)
{
    cout << "Loading texture: " << sTextureFile << "..." << endl;

    auto it = textures.find(sTextureFile);

    if(it != textures.end())
    {
        cout << sTextureFile << " was cached" << endl;
        return it->second;
    }

    cout << sTextureFile << " was not cached - loading." << endl;

    pvr_ptr_t tex;
    load_texture(&tex, 512 * 512 * 2, sTextureFile.c_str());
    textures[sTextureFile] = tex;

    cout << sTextureFile << " loaded" << endl;
    return static_cast<void *>(tex);
}

void GameJamGame::Reload()
{
    titleScreen = LoadTexture("rd/title.png", 512, 512);

    camera = Camera(Vector(0, 6, -5));

    ross = Player();
    ross.SetStaticMesh(LoadStaticMesh("ross", "rd/ross.ross_mesh", "rd/ross.png", 512, 512));

    dungeon = LoadStaticMesh("dungeon", "rd/dungeon.ross_mesh", "rd/dungeon.png", 512, 512);

    enemies.clear();
    size_t numEnemies = 3;

    StaticMesh goblin = LoadStaticMesh("goblin", "rd/goblin.ross_mesh", "rd/goblin.png", 512, 512);

    for(size_t i = 0; i < numEnemies; ++i)
    {
        Enemy enemy;

        enemy.SetStaticMesh(goblin);

        float x = ((rand() / (float) RAND_MAX) * fHalfDungeonWidth * 2) - fHalfDungeonWidth;
        float y = ((rand() / (float) RAND_MAX) * fHalfDungeonHeight * 2) - fHalfDungeonHeight;

        Vector pos(x, 0, y);
        enemy.SetPosition(pos);

        enemies.push_back(enemy);
    }
}

GameJamGame::~GameJamGame()
{
}

Vector DungeonCollisionResponse(Vector const & input)
{
    Vector pos = input;

    if(pos.x < -fHalfDungeonWidth)
    {
        pos.x = -fHalfDungeonWidth;
    }
    else if(pos.x > fHalfDungeonWidth)
    {
        pos.x = fHalfDungeonWidth;
    }
    
    if(pos.z < -fHalfDungeonHeight)
    {
        pos.z = -fHalfDungeonHeight;
    }
    else if(pos.z > fHalfDungeonHeight)
    {
        pos.z = fHalfDungeonHeight;
    }

    return pos;
}

void GameJamGame::Process(float deltaTime, Input * pInput)
{
    cont_state_t * st = pInput->pControllerState;

    if(st != NULL)
    {
        static bool startPressed;

        if(st->buttons & CONT_START)
        {
            if(startPressed == false)
            {
                startPressed = true;
            }
        }
        else if(startPressed == true)
        {
            Reload();
            state = (state == Gameplay) ? TitleScreen : Gameplay;
            startPressed = false;
        }
    }

    if(state == Gameplay)
    {   
        camera.Process(deltaTime, pInput);
        dungeon.Process(deltaTime, pInput);

        ross.Process(deltaTime, pInput);

        for(size_t i = 0; i < enemies.size(); ++i)
        {
            Enemy & enemy = enemies.at(i);
            enemy.SetTarget(&ross);
            enemy.Process(deltaTime, pInput);
        }

        /**
        * Collision response.
        */

        {
            ross.SetPosition(DungeonCollisionResponse(ross.GetPosition()));

            for(size_t i = 0; i < enemies.size(); ++i)
            {
                Enemy & enemy = enemies.at(i);
                enemy.SetPosition(DungeonCollisionResponse(enemy.GetPosition()));
            }
        }

        camera.SetTarget(ross.GetPosition());
   }
}

void GameJamGame::Draw()
{
    if(state == TitleScreen)
    {
        static float const w = 640;
        static float const h = 480;

        static tQuads quads {

            Quad(Vertex(0, 0, 1, 0, 0),
                 Vertex(w, 0, 1, 1, 0),
                 Vertex(0, h, 1, 0, 1),
                 Vertex(w, h, 1, 1, 1))
        };

        render::begin_frame();
        pvr_list_begin(PVR_LIST_OP_POLY);
        render::draw_textured_quads(quads, titleScreen, 512, 512);
        pvr_list_finish();
        render::end_frame();
    }
    else
    {
        /**
         * Apply the current view projection matrix.
         */

        Matrix const & vp = camera.GetViewProjection();
        dungeon.SetTransform(vp);
        ross.SetViewProjection(vp);

        for(size_t i = 0; i < enemies.size(); ++i)
        {
            Enemy & enemy = enemies.at(i);
            enemy.SetViewProjection(vp);
        }

        /**
         * Draw entities.
         */

        render::begin_frame();
        pvr_list_begin(PVR_LIST_OP_POLY);

        dungeon.Draw();

        ross.Draw();

        for(size_t i = 0; i < enemies.size(); ++i)
        {
            Enemy & enemy = enemies.at(i);
            enemy.Draw();
        }

        pvr_list_finish();
        render::end_frame();
    }
}