#include "math_types.hpp"

#include <iostream>
using std::cout;
using std::endl;

void CalculateInverseBindPose(Joint * _pJoint)
{
    /**
     * Calculate inverse bind pose matrix for each joint of the
     * bind pose.
     */

    Joint * pBindPoseJoint = _pJoint;

    while(pBindPoseJoint != NULL)
    {
        Matrix inverseBindPose;

        {
            inverseBindPose.LoadIdentity();

            Joint * pJoint = pBindPoseJoint;

            while(pJoint != NULL)
            {
                float s = pJoint->scale;

                if(s != 0)
                {
                    s = 1.f / s;
                }

                Matrix scaleMatrix(s, 0, 0, 0,
                                    0, s, 0, 0,
                                    0, 0, s, 0,
                                    0, 0, 0, 1);

                Matrix rotMatrix = pJoint->rot.ToMatrix();
                rotMatrix.Transpose();

                Vector loc = pJoint->loc;
                loc.x *= -1;
                loc.y *= -1;
                loc.z *= -1;

                Matrix locMatrix;
                locMatrix.LoadIdentity();
                locMatrix.Translate(loc);

                pJoint = pJoint->parent;

                // Apply to inverseBindPose.

                Matrix jointTrans;

                jointTrans.LoadIdentity();

                jointTrans.Mult(scaleMatrix);
                jointTrans.Mult(locMatrix);
                jointTrans.Mult(rotMatrix);
                jointTrans.Mult(inverseBindPose);

                inverseBindPose = jointTrans;
            }
        }

        pBindPoseJoint->pose = inverseBindPose;

        pBindPoseJoint = pBindPoseJoint->GetChild(0);
    }
}

void CalculateForwardPose(Joint * _pJoint)
{
     /**
      * Calculate pose matrix for each joint of the current pose.
      * (very inefficient, could make this O(n) instead of O(n*n))
      */

    Joint * pDynamicPoseJoint = _pJoint;

    while(pDynamicPoseJoint != NULL)
    {
        Matrix pose;

        {
            pose.LoadIdentity();

            Joint * pJoint = _pJoint;

            while(pJoint != NULL)
            {
                float s = pJoint->scale;

                Matrix scaleMatrix(s, 0, 0, 0,
                                    0, s, 0, 0,
                                    0, 0, s, 0,
                                    0, 0, 0, 1);

                Matrix rotMatrix = pJoint->rot.ToMatrix();

                Vector loc = pJoint->loc;

                Matrix locMatrix(1, 0, 0, 0,
                                    0, 1, 0, 0,
                                    0, 0, 1, 0,
                                    loc.x, loc.y, loc.z, 1);

                if(pJoint == pDynamicPoseJoint)
                {
                    pJoint = NULL;
                }
                else
                {
                    pJoint = pJoint->GetChild(0);
                }

                Matrix jointTrans;
                jointTrans.LoadIdentity();

                jointTrans.Mult(rotMatrix);
                jointTrans.Mult(locMatrix);
                jointTrans.Mult(scaleMatrix);
                jointTrans.Mult(pose);

                pose = jointTrans;
            }
        }

        pDynamicPoseJoint->pose = pose;
        pDynamicPoseJoint = pDynamicPoseJoint->GetChild(0);
    }
}

Joint * GetJointByID(Joint * pRoot, int id)
{
    if(pRoot->id == id)
    {
        return pRoot;
    }

    for(size_t i = 0; i < pRoot->children.size(); ++i)
    {
        Joint * pResult = GetJointByID(pRoot->GetChild(i), id);

        if(pResult)
        {
            return pResult;
        }
    }

    return NULL;
}

void print_matrix(Matrix const & m)
{
    cout << m.m00 << ", " << m.m01 << ", " << m.m02 << ", " << m.m03 << "\n";
    cout << m.m10 << ", " << m.m11 << ", " << m.m12 << ", " << m.m13 << "\n";
    cout << m.m20 << ", " << m.m21 << ", " << m.m22 << ", " << m.m23 << "\n";
    cout << m.m30 << ", " << m.m31 << ", " << m.m32 << ", " << m.m33 << "\n";
    cout << endl;
}

void print_vertex_weights(Weights const & w)
{
    cout << "Joint " << w.w0.first << ",  weight: " << w.w0.second << endl;
    cout << "Joint " << w.w1.first << ",  weight: " << w.w1.second << endl;
    cout << "Joint " << w.w2.first << ",  weight: " << w.w2.second << endl;
}

void Transform(Vector & vec, Matrix & mat)
{
    Vector result(0, 0, 0, 0);

    result.x = vec.x * mat.m00 + vec.y * mat.m10 + vec.z * mat.m20 + vec.w * mat.m30;
    result.y = vec.x * mat.m01 + vec.y * mat.m11 + vec.z * mat.m21 + vec.w * mat.m31;
    result.z = vec.x * mat.m02 + vec.y * mat.m12 + vec.z * mat.m22 + vec.w * mat.m32;
    result.w = vec.x * mat.m03 + vec.y * mat.m13 + vec.z * mat.m23 + vec.w * mat.m33;

    vec = result;
}

typedef std::vector<Quad> tQuads;
typedef std::vector<Joint> tJoints;