#pragma once

#include <cmath>
#include <vector>

float const pi = 3.1415926535f;

class Rect
{
public:

    Rect(float left,
         float right,
         float top,
         float bottom,
         float depth = 1)

    : left (left)
    , right (right)
    , top (top)
    , bottom (bottom)
    , depth (depth)
    {}

    float left, right;
    float top, bottom;
    float depth;
};

typedef std::pair<int, float> tWeight;

class Weights
{
public:

    Weights()
    : w0 (tWeight(0, 0))
    , w1 (tWeight(0, 0))
    , w2 (tWeight(0, 0))
    {}

    Weights(tWeight w0, tWeight w1, tWeight w2)
    : w0 (w0)
    , w1 (w1)
    , w2 (w2)
    {}

    tWeight w0;
    tWeight w1;
    tWeight w2;
};

class Vertex
{
public:

    Vertex()
    : x (0)
    , y (0)
    , z (0)
    , u (0)
    , v (0)
    , r (1)
    , g (1)
    , b (1)
    , a (1)
    , weights ()
    {}

    Vertex(float x,
           float y,
           float z = 0,
           float u = 0,
           float v = 0,
           float r = 1,
           float g = 1,
           float b = 1,
           float a = 1,
           Weights w = Weights())

    : x (x)
    , y (y)
    , z (z)
    , u (u)
    , v (v)
    , r (r)
    , g (g)
    , b (b)
    , a (a)
    , weights (w)
    {}

    float x, y, z;
    float u, v;
    float r, g, b, a;
    Weights weights;
};

class Quad
{
public:

    Quad(Rect const & r)
    : v0 (Vertex(r.left, r.top, r.depth, 0, 0))
    , v1 (Vertex(r.right, r.top, r.depth, 1, 0))
    , v2 (Vertex(r.left, r.bottom, r.depth, 0, 1))
    , v3 (Vertex(r.right, r.bottom, r.depth, 1, 1))
    {}

    Quad(Vertex const & v0,
         Vertex const & v1,
         Vertex const & v2,
         Vertex const & v3)

    : v0 (v0)
    , v1 (v1)
    , v2 (v2)
    , v3 (v3)
    {}

    Vertex v0, v1, v2, v3;
};

class Tri
{
public:

    Tri(Vertex const & v0,
        Vertex const & v1,
        Vertex const & v2)

    : v0 (v0)
    , v1 (v1)
    , v2 (v2)
    {}

    Vertex v0, v1, v2;
};

class Vector
{
public:

    Vector()
    : x ()
    , y ()
    , z ()
    , w ()
    {}

    Vector(float x, float y, float z = 1, float w = 1)
    : x (x)
    , y (y)
    , z (z)
    , w (w)
    {}

    void Add(Vector & v)
    {
        x += v.x;
        y += v.y;
        z += v.z;
    }

    void Sub(Vector & v)
    {
        x -= v.x;
        y -= v.y;
        z -= v.z;
    }

    void Scale(float scale)
    {
        x *= scale;
        y *= scale;
        z *= scale;
    }

    void Dehomogenize()
    {
        if(w != 0)
        {
            x /= w;
            y /= w;
            z /= w;
            w /= w;
        }
    }

    static Vector Sub(Vector const & v1, Vector const & v2)
    {
        return Vector(v1.x - v2.x, v1.y - v2.y, v1.z - v2.z, v1.w);
    }

    static float Length(Vector const & v)
    {
        return sqrt(v.x * v.x + v.y * v.y + v.z * v.z);
    }

    static Vector Normalize(Vector const & v)
    {
        float len = Length(v);

        if(len == 0)
        {
            len = 0.0001f;
        }

        return Vector(v.x / len, v.y / len, v.z / len, v.w);
    }

    static Vector CrossProduct(Vector const & v1, Vector const & v2)
    {
        Vector result(v1);

        result.x = v1.y * v2.z - v1.z * v2.y;
        result.y = v1.z * v2.x - v1.x * v2.z;
        result.z = v1.x * v2.y - v1.y * v2.x;

        return result;
    }

    static float DotProduct(Vector const & v1, Vector const & v2)
    {
        return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z; 
    }

    float x, y, z, w;
};

class Matrix
{
public:

    Matrix()
    : m00 (0), m01 (0), m02 (0), m03 (0)
    , m10 (0), m11 (0), m12 (0), m13 (0)
    , m20 (0), m21 (0), m22 (0), m23 (0)
    , m30 (0), m31 (0), m32 (0), m33 (0)
    {}

    Matrix(float m00, float m01, float m02, float m03,
           float m10, float m11, float m12, float m13,
           float m20, float m21, float m22, float m23,
           float m30, float m31, float m32, float m33)

    : m00 (m00), m01 (m01), m02 (m02), m03 (m03)
    , m10 (m10), m11 (m11), m12 (m12), m13 (m13)
    , m20 (m20), m21 (m21), m22 (m22), m23 (m23)
    , m30 (m30), m31 (m31), m32 (m32), m33 (m33)
    {}

    void LoadIdentity()
    {
        *this = Matrix();
        m00 = 1;
        m11 = 1;
        m22 = 1;
        m33 = 1;
    }

    void Translate(Vector const & translation)
    {
        m30 += translation.x;
        m31 += translation.y;
        m32 += translation.z;
    }

    void Mult(Matrix const & m)
    {
        Matrix result;

        result.m00 = m00 * m.m00 + m01 * m.m10 + m02 * m.m20 + m03 * m.m30;
        result.m01 = m00 * m.m01 + m01 * m.m11 + m02 * m.m21 + m03 * m.m31;
        result.m02 = m00 * m.m02 + m01 * m.m12 + m02 * m.m22 + m03 * m.m32;
        result.m03 = m00 * m.m03 + m01 * m.m13 + m02 * m.m23 + m03 * m.m33;

        result.m10 = m10 * m.m00 + m11 * m.m10 + m12 * m.m20 + m13 * m.m30;
        result.m11 = m10 * m.m01 + m11 * m.m11 + m12 * m.m21 + m13 * m.m31;
        result.m12 = m10 * m.m02 + m11 * m.m12 + m12 * m.m22 + m13 * m.m32;
        result.m13 = m10 * m.m03 + m11 * m.m13 + m12 * m.m23 + m13 * m.m33;

        result.m20 = m20 * m.m00 + m21 * m.m10 + m22 * m.m20 + m23 * m.m30;
        result.m21 = m20 * m.m01 + m21 * m.m11 + m22 * m.m21 + m23 * m.m31;
        result.m22 = m20 * m.m02 + m21 * m.m12 + m22 * m.m22 + m23 * m.m32;
        result.m23 = m20 * m.m03 + m21 * m.m13 + m22 * m.m23 + m23 * m.m33;

        result.m30 = m30 * m.m00 + m31 * m.m10 + m32 * m.m20 + m33 * m.m30;
        result.m31 = m30 * m.m01 + m31 * m.m11 + m32 * m.m21 + m33 * m.m31;
        result.m32 = m30 * m.m02 + m31 * m.m12 + m32 * m.m22 + m33 * m.m32;
        result.m33 = m30 * m.m03 + m31 * m.m13 + m32 * m.m23 + m33 * m.m33;

        *this = result;
    }

    void Transpose()
    {
        Matrix result;

        result.m00 = m00;
        result.m01 = m10;
        result.m02 = m20;
        result.m03 = m30;

        result.m10 = m01;
        result.m11 = m11;
        result.m12 = m21;
        result.m13 = m31;

        result.m20 = m02;
        result.m21 = m12;
        result.m22 = m22;
        result.m23 = m32;

        result.m30 = m03;
        result.m31 = m13;
        result.m32 = m23;
        result.m33 = m33;

        *this = result;
    }

    void LookAt(Vector const & cameraPosition, Vector const & cameraTarget, Vector const & cameraUp)
    {
        Vector zaxis = Vector::Normalize(Vector::Sub(cameraTarget, cameraPosition));
        Vector xaxis = Vector::Normalize(Vector::CrossProduct(cameraUp, zaxis));
        Vector yaxis = Vector::CrossProduct(zaxis, xaxis);

        m00 = xaxis.x;
        m01 = yaxis.x;
        m02 = zaxis.x;
        m03 = 0;

        m10 = xaxis.y;
        m11 = yaxis.y;
        m12 = zaxis.y;
        m13 = 0;

        m20 = xaxis.z;
        m21 = yaxis.z;
        m22 = zaxis.z;
        m23 = 0;

        m30 = -Vector::DotProduct(xaxis, cameraPosition);
        m21 = -Vector::DotProduct(yaxis, cameraPosition);
        m22 = -Vector::DotProduct(zaxis, cameraPosition);
        m23 = 1;
    }

    void Perspective(float fov, float aspect, float near, float far, bool bLeftHanded)
    {
        if(fov <= 0 || aspect == 0)
        {
            LoadIdentity();
            return;
        }

        float frustumDepth = far - near;
        float oneOverDepth = 1 / frustumDepth;

        m11 = 1.f / tan(0.5f * fov);
        m00 = (bLeftHanded ? 1 : -1) * m11 / aspect;
        m22 = (far + near) * oneOverDepth;
        m32 = -(2 * near * far) * oneOverDepth;
        m23 = 1;
        m33 = 0;
    }

    float m00, m01, m02, m03;
    float m10, m11, m12, m13;
    float m20, m21, m22, m23;
    float m30, m31, m32, m33;
};

class Quaternion
{
public:

    Quaternion()
    : x ()
    , y ()
    , z ()
    , w ()
    {}

    Quaternion(float x, float y, float z, float w)
    : x (x)
    , y (y)
    , z (z)
    , w (w)
    {}

    Quaternion(Vector const & axis, float angle)
    {
        double const a_by_2 = angle / 2.f;
        double s = sin(a_by_2);
        x = axis.x * s;
        y = axis.y * s;
        z = axis.z * s;
        w = cos(a_by_2);
    }

    void ToAxisAngle(Vector & axis, float & angle)
    {
        Quaternion q = *this;

        // if w>1 acos and sqrt will produce errors, this can't happen if quaternion is normalised
        if (q.w > 1)
        {
            q.Normalize();
        }
        
        angle = 2 * acos(q.w);

        double s = sqrt(1 - q.w * q.w); // assuming quaternion normalised then w is less than 1, so term always positive.
        
        if (s < 0.001)
        {  
            // test to avoid divide by zero, s is always positive due to sqrt
            // if s close to zero then direction of axis not important
            axis.x = q.x; // if it is important that axis is normalised then replace with x=1; y=z=0;
            axis.y = q.y;
            axis.z = q.z;
        }
        else
        {
            axis.x = q.x / s; // normalise axis
            axis.y = q.y / s;
            axis.z = q.z / s;
        }
    }

    void Normalize()
    {
        double mag = sqrt(x * x + y * y + z * z + w * w);

        if(mag != 0)
        {
            x /= mag;
            y /= mag;
            z /= mag;
            w /= mag;
        }
    }

    Matrix ToMatrix()
    {
        Matrix result;

        double xx = x * x;
        double xy = x * y;
        double xz = x * z;
        double xw = x * w;

        double yy = y * y;
        double yz = y * z;
        double yw = y * w;

        double zz = z * z;
        double zw = z * w;

        result.m00 = 1 - 2 * (yy + zz);
        result.m01 = 2 * (xy - zw);
        result.m02 = 2 * (xz + yw);

        result.m10 = 2 * (xy + zw);
        result.m11 = 1 - 2 * (xx + zz);
        result.m12 = 2 * (yz - xw);

        result.m20 = 2 * (xz - yw);
        result.m21 = 2 * (yz + xw);
        result.m22 = 1 - 2 * (xx + yy);

        result.m33 = 1;

        return result;
    }

    float x, y, z, w;
};

class Joint
{
public:

    Joint(int id, Vector loc, Quaternion rot, float scale)
    : id (id)
    , loc (loc)
    , rot (rot)
    , scale (scale)
    , parent (NULL)
    {}

    Joint(Joint const & j, Joint * parent)
    {
        *this = j;

        if(this->parent != parent)
        {
            this->parent = parent;
            this->parent->children.push_back(this);
        }
    }

    Joint * GetChild(size_t childIndex)
    {
        if(childIndex >= 0 && childIndex < children.size())
        {
            return children[childIndex];
        }
        else
        {
            return NULL;
        }
    }

    int id;
    Vector loc;
    Quaternion rot;
    float scale;
    Matrix pose;
    Joint * parent;
    std::vector<Joint*> children;
};

void Transform(Vector & vec, Matrix & mat);

void CalculateInverseBindPose(Joint * _pJoint);

void CalculateForwardPose(Joint * _pJoint);

Joint * GetJointByID(Joint * pRoot, int id);

void print_matrix(Matrix const & m);

void print_vertex_weights(Weights const & w);

typedef std::vector<Quad> tQuads;
typedef std::vector<Tri> tTris;
typedef std::vector<Joint> tJoints;